// Copyright (c) Tailscale Inc & AUTHORS
// SPDX-License-Identifier: BSD-3-Clause

package appc

import (
	"context"
	"net/netip"
	"reflect"
	"slices"
	"testing"
	"time"

	xmaps "golang.org/x/exp/maps"
	"golang.org/x/net/dns/dnsmessage"
	"tailscale.com/appc/appctest"
	"tailscale.com/appc/routeinfo"
	"tailscale.com/util/mak"
	"tailscale.com/util/must"
)

func TestUpdateDomainsWithPresistStore(t *testing.T) {

	ctx := context.Background()
	a := NewAppConnector(t.Logf, nil, true)
	TestRouteAdvertiser := &appctest.RouteCollector{}
	a.routeAdvertiser = TestRouteAdvertiser
	prefixControl := netip.MustParsePrefix("192.0.0.8/32")
	prefixLocal := netip.MustParsePrefix("192.0.0.9/32")
	prefixDiscovered := netip.MustParsePrefix("192.0.0.10/32")
	prefixShouldUnadvertise := netip.MustParsePrefix("192.0.0.11/32")
	ri := routeinfo.NewRouteInfo()
	ri.Control = []netip.Prefix{prefixControl}
	ri.Local = []netip.Prefix{prefixLocal}
	a.UpdateRouteInfo(ri)
	a.UpdateDomains([]string{"example.com", "test.com"})

	a.Wait(ctx)
	if got, want := a.Domains().AsSlice(), []string{"example.com", "test.com"}; !slices.Equal(got, want) {
		t.Errorf("got %v; want %v", got, want)
	}
	if _, hasKey := a.routeInfo.Discovered["example.com"]; !hasKey {
		t.Errorf("Discovered did not record the test domain.")
	}

	now := time.Now()
	testRoutes := make(map[netip.Prefix]time.Time)
	testRoutes[prefixDiscovered] = now
	exampleRoutes := make(map[netip.Prefix]time.Time)
	exampleRoutes[prefixShouldUnadvertise] = now
	exampleRoutes[prefixControl] = now
	exampleRoutes[prefixLocal] = now
	exampleRoutes[prefixDiscovered] = now
	ri.Discovered["test.com"] = &routeinfo.DatedRoutes{Routes: testRoutes, LastCleanup: now}
	ri.Discovered["example.com"] = &routeinfo.DatedRoutes{Routes: exampleRoutes, LastCleanup: now}
	a.UpdateRouteInfo(ri)
	TestRouteAdvertiser.SetRoutes(ri.Discovered["example.com"].RoutesSlice())

	addrControl := netip.MustParseAddr("192.0.0.8")
	addrLocal := netip.MustParseAddr("192.0.0.9")
	addrDiscovered := netip.MustParseAddr("192.0.0.10")
	addrShouldUnadvertise := netip.MustParseAddr("192.0.0.11")
	a.domains["example.com"] = append(a.domains["example.com"], []netip.Addr{addrControl, addrLocal, addrDiscovered, addrShouldUnadvertise}...)
	a.domains["test.com"] = append(a.domains["test.com"], addrDiscovered)
	a.UpdateDomains([]string{"example.com", "test.com"})
	a.Wait(ctx)

	if got, want := a.domains["example.com"], []netip.Addr{addrControl, addrLocal, addrDiscovered, addrShouldUnadvertise}; !slices.Equal(got, want) {
		t.Errorf("got %v; want %v", got, want)
	}

	// domains are explicitly downcased on set.
	a.UpdateDomains([]string{"UP.EXAMPLE.COM", "test.com"})
	a.Wait(ctx)
	if got, want := xmaps.Keys(a.domains), []string{"up.example.com", "test.com"}; !slices.Equal(got, want) {
		t.Errorf("got %v; want %v", got, want)
	}

	if got, want := TestRouteAdvertiser.RemovedRoutes(), []netip.Prefix{prefixShouldUnadvertise}; !slices.Equal(got, want) {
		t.Errorf("got %v; want %v", got, want)
	}

	if got, want := TestRouteAdvertiser.Routes(), []netip.Prefix{prefixControl, prefixLocal, prefixDiscovered}; !slices.Equal(got, want) {
		t.Errorf("got %v; want %v", got, want)
	}

}

func TestUpdateDomainsWithoutPresistStore(t *testing.T) {
	ctx := context.Background()
	a := NewAppConnector(t.Logf, nil, false)
	a.routeAdvertiser = &appctest.RouteCollector{}
	a.UpdateRouteInfo(routeinfo.NewRouteInfo())
	a.UpdateDomains([]string{"example.com"})

	a.Wait(ctx)
	if got, want := a.Domains().AsSlice(), []string{"example.com"}; !slices.Equal(got, want) {
		t.Errorf("got %v; want %v", got, want)
	}

	addr := netip.MustParseAddr("192.0.0.8")
	a.domains["example.com"] = append(a.domains["example.com"], addr)
	a.UpdateDomains([]string{"example.com"})
	a.Wait(ctx)

	if got, want := a.domains["example.com"], []netip.Addr{addr}; !slices.Equal(got, want) {
		t.Errorf("got %v; want %v", got, want)
	}

	// domains are explicitly downcased on set.
	a.UpdateDomains([]string{"UP.EXAMPLE.COM"})
	a.Wait(ctx)
	if got, want := xmaps.Keys(a.domains), []string{"up.example.com"}; !slices.Equal(got, want) {
		t.Errorf("got %v; want %v", got, want)
	}
}

func TestUpdateRoutes(t *testing.T) {
	for _, shouldStore := range []bool{true, false} {
		ctx := context.Background()
		rc := &appctest.RouteCollector{}
		a := NewAppConnector(t.Logf, rc, shouldStore)
		a.updateDomains([]string{"*.example.com"})

		// This route should be collapsed into the range
		a.ObserveDNSResponse(dnsResponse("a.example.com.", "192.0.2.1"))
		a.Wait(ctx)

		if !slices.Equal(rc.Routes(), []netip.Prefix{netip.MustParsePrefix("192.0.2.1/32")}) {
			t.Fatalf("got %v, want %v", rc.Routes(), []netip.Prefix{netip.MustParsePrefix("192.0.2.1/32")})
		}

		// This route should not be collapsed or removed
		a.ObserveDNSResponse(dnsResponse("b.example.com.", "192.0.0.1"))
		a.Wait(ctx)

		routes := []netip.Prefix{netip.MustParsePrefix("192.0.2.0/24"), netip.MustParsePrefix("192.0.0.1/32")}
		a.updateRoutes(routes)

		if shouldStore {
			wantRouteInfoControlRoutes := routes
			if !slices.Equal(a.routeInfo.Control, wantRouteInfoControlRoutes) {
				t.Fatalf("got %v, want %v (shouldStore=%t)", a.routeInfo.Control, wantRouteInfoControlRoutes, shouldStore)
			}
		} else {
			if a.routeInfo != nil {
				t.Fatalf("got %v, want %v (shouldStore=%t)", a.routeInfo.Control, nil, shouldStore)
			}
		}

		slices.SortFunc(rc.Routes(), prefixCompare)
		rc.SetRoutes(slices.Compact(rc.Routes()))
		slices.SortFunc(routes, prefixCompare)

		// Ensure that the non-matching /32 is preserved, even though it's in the domains table.
		if !slices.EqualFunc(routes, rc.Routes(), prefixEqual) {
			t.Errorf("added routes: got %v, want %v", rc.Routes(), routes)
		}

		// Ensure that the contained /32 is removed, replaced by the /24.
		wantRemoved := []netip.Prefix{netip.MustParsePrefix("192.0.2.1/32")}
		if !slices.EqualFunc(rc.RemovedRoutes(), wantRemoved, prefixEqual) {
			t.Fatalf("unexpected removed routes: %v", rc.RemovedRoutes())
		}
	}
}

func TestUpdateRoutesNotUnadvertiseRoutesFromOtherSources(t *testing.T) {
	for _, shouldStore := range []bool{true, false} {
		rc := &appctest.RouteCollector{}
		a := NewAppConnector(t.Logf, rc, shouldStore)
		testRi := routeinfo.NewRouteInfo()
		testRi.Local = append(testRi.Local, netip.MustParsePrefix("192.0.2.0/24"))
		a.routeInfo = testRi
		rc.StoreRouteInfo(testRi)

		routes := []netip.Prefix{netip.MustParsePrefix("192.0.2.0/24"), netip.MustParsePrefix("192.0.0.1/32")}
		a.updateRoutes(routes)
		wantRouteInfoControlRoutes := []netip.Prefix{}
		if shouldStore {
			wantRouteInfoControlRoutes = routes
		}
		if !slices.Equal(a.routeInfo.Control, wantRouteInfoControlRoutes) {
			t.Fatalf("got %v, want %v (shouldStore=%t)", a.routeInfo.Control, wantRouteInfoControlRoutes, shouldStore)
		}

		routes2 := []netip.Prefix{netip.MustParsePrefix("192.0.0.1/32")}
		a.updateRoutes(routes2)

		wantRemoved := []netip.Prefix{}
		if !slices.EqualFunc(rc.RemovedRoutes(), wantRemoved, prefixEqual) {
			t.Fatalf("unexpected removed routes: %v", rc.RemovedRoutes())
		}
	}
}

func TestUpdateRoutesUnadvertisesContainedRoutes(t *testing.T) {
	for _, shouldStore := range []bool{true, false} {
		rc := &appctest.RouteCollector{}
		a := NewAppConnector(t.Logf, rc, shouldStore)
		mak.Set(&a.domains, "example.com", []netip.Addr{netip.MustParseAddr("192.0.2.1")})
		rc.SetRoutes([]netip.Prefix{netip.MustParsePrefix("192.0.2.1/32")})
		routes := []netip.Prefix{netip.MustParsePrefix("192.0.2.0/24")}
		a.updateRoutes(routes)

		if !slices.EqualFunc(routes, rc.Routes(), prefixEqual) {
			t.Fatalf("got %v, want %v", rc.Routes(), routes)
		}
	}
}

func TestDomainRoutes(t *testing.T) {
	for _, shouldStore := range []bool{true, false} {
		rc := &appctest.RouteCollector{}
		a := NewAppConnector(t.Logf, rc, shouldStore)
		a.updateDomains([]string{"example.com"})
		a.ObserveDNSResponse(dnsResponse("example.com.", "192.0.0.8"))
		a.Wait(context.Background())

		want := map[string][]netip.Addr{
			"example.com": {netip.MustParseAddr("192.0.0.8")},
		}

		if got := a.DomainRoutes(); !reflect.DeepEqual(got, want) {
			t.Fatalf("DomainRoutes: got %v, want %v", got, want)
		}
	}
}

func TestObserveDNSResponse(t *testing.T) {
	for _, shouldStore := range []bool{true, false} {
		ctx := context.Background()
		rc := &appctest.RouteCollector{}
		a := NewAppConnector(t.Logf, rc, shouldStore)

		// a has no domains configured, so it should not advertise any routes
		a.ObserveDNSResponse(dnsResponse("example.com.", "192.0.0.8"))
		if got, want := rc.Routes(), ([]netip.Prefix)(nil); !slices.Equal(got, want) {
			t.Errorf("got %v; want %v", got, want)
		}

		wantRoutes := []netip.Prefix{netip.MustParsePrefix("192.0.0.8/32")}

		a.updateDomains([]string{"example.com"})
		a.ObserveDNSResponse(dnsResponse("example.com.", "192.0.0.8"))
		a.Wait(ctx)
		if got, want := rc.Routes(), wantRoutes; !slices.Equal(got, want) {
			t.Errorf("got %v; want %v", got, want)
		}

		// a CNAME record chain should result in a route being added if the chain
		// matches a routed domain.
		a.updateDomains([]string{"www.example.com", "example.com"})
		a.ObserveDNSResponse(dnsCNAMEResponse("192.0.0.9", "www.example.com.", "chain.example.com.", "example.com."))
		a.Wait(ctx)
		wantRoutes = append(wantRoutes, netip.MustParsePrefix("192.0.0.9/32"))
		if got, want := rc.Routes(), wantRoutes; !slices.Equal(got, want) {
			t.Errorf("got %v; want %v", got, want)
		}

		// a CNAME record chain should result in a route being added if the chain
		// even if only found in the middle of the chain
		a.ObserveDNSResponse(dnsCNAMEResponse("192.0.0.10", "outside.example.org.", "www.example.com.", "example.org."))
		a.Wait(ctx)
		wantRoutes = append(wantRoutes, netip.MustParsePrefix("192.0.0.10/32"))
		if got, want := rc.Routes(), wantRoutes; !slices.Equal(got, want) {
			t.Errorf("got %v; want %v", got, want)
		}

		wantRoutes = append(wantRoutes, netip.MustParsePrefix("2001:db8::1/128"))

		a.ObserveDNSResponse(dnsResponse("example.com.", "2001:db8::1"))
		a.Wait(ctx)
		if got, want := rc.Routes(), wantRoutes; !slices.Equal(got, want) {
			t.Errorf("got %v; want %v", got, want)
		}

		// don't re-advertise routes that have already been advertised
		a.ObserveDNSResponse(dnsResponse("example.com.", "2001:db8::1"))
		a.Wait(ctx)
		if !slices.Equal(rc.Routes(), wantRoutes) {
			t.Errorf("rc.Routes(): got %v; want %v", rc.Routes(), wantRoutes)
		}

		// don't advertise addresses that are already in a control provided route
		pfx := netip.MustParsePrefix("192.0.2.0/24")
		a.updateRoutes([]netip.Prefix{pfx})
		wantRoutes = append(wantRoutes, pfx)
		a.ObserveDNSResponse(dnsResponse("example.com.", "192.0.2.1"))
		a.Wait(ctx)
		if !slices.Equal(rc.Routes(), wantRoutes) {
			t.Errorf("rc.Routes(): got %v; want %v", rc.Routes(), wantRoutes)
		}
		if !slices.Contains(a.domains["example.com"], netip.MustParseAddr("192.0.2.1")) {
			t.Errorf("missing %v from %v", "192.0.2.1", a.domains["exmaple.com"])
		}
	}
}

func TestWildcardDomains(t *testing.T) {
	for _, shouldStore := range []bool{true, false} {
		ctx := context.Background()
		rc := &appctest.RouteCollector{}
		a := NewAppConnector(t.Logf, rc, shouldStore)

		a.updateDomains([]string{"*.example.com"})
		a.ObserveDNSResponse(dnsResponse("foo.example.com.", "192.0.0.8"))
		a.Wait(ctx)
		if got, want := rc.Routes(), []netip.Prefix{netip.MustParsePrefix("192.0.0.8/32")}; !slices.Equal(got, want) {
			t.Errorf("routes: got %v; want %v", got, want)
		}
		if got, want := a.wildcards, []string{"example.com"}; !slices.Equal(got, want) {
			t.Errorf("wildcards: got %v; want %v", got, want)
		}

		a.updateDomains([]string{"*.example.com", "example.com"})
		if _, ok := a.domains["foo.example.com"]; !ok {
			t.Errorf("expected foo.example.com to be preserved in domains due to wildcard")
		}
		if got, want := a.wildcards, []string{"example.com"}; !slices.Equal(got, want) {
			t.Errorf("wildcards: got %v; want %v", got, want)
		}

		// There was an early regression where the wildcard domain was added repeatedly, this guards against that.
		a.updateDomains([]string{"*.example.com", "example.com"})
		if len(a.wildcards) != 1 {
			t.Errorf("expected only one wildcard domain, got %v", a.wildcards)
		}
	}
}

// dnsResponse is a test helper that creates a DNS response buffer for the given domain and address
func dnsResponse(domain, address string) []byte {
	addr := netip.MustParseAddr(address)
	b := dnsmessage.NewBuilder(nil, dnsmessage.Header{})
	b.EnableCompression()
	b.StartAnswers()
	switch addr.BitLen() {
	case 32:
		b.AResource(
			dnsmessage.ResourceHeader{
				Name:  dnsmessage.MustNewName(domain),
				Type:  dnsmessage.TypeA,
				Class: dnsmessage.ClassINET,
				TTL:   0,
			},
			dnsmessage.AResource{
				A: addr.As4(),
			},
		)
	case 128:
		b.AAAAResource(
			dnsmessage.ResourceHeader{
				Name:  dnsmessage.MustNewName(domain),
				Type:  dnsmessage.TypeAAAA,
				Class: dnsmessage.ClassINET,
				TTL:   0,
			},
			dnsmessage.AAAAResource{
				AAAA: addr.As16(),
			},
		)
	default:
		panic("invalid address length")
	}
	return must.Get(b.Finish())
}

func dnsCNAMEResponse(address string, domains ...string) []byte {
	addr := netip.MustParseAddr(address)
	b := dnsmessage.NewBuilder(nil, dnsmessage.Header{})
	b.EnableCompression()
	b.StartAnswers()

	if len(domains) >= 2 {
		for i, domain := range domains[:len(domains)-1] {
			b.CNAMEResource(
				dnsmessage.ResourceHeader{
					Name:  dnsmessage.MustNewName(domain),
					Type:  dnsmessage.TypeCNAME,
					Class: dnsmessage.ClassINET,
					TTL:   0,
				},
				dnsmessage.CNAMEResource{
					CNAME: dnsmessage.MustNewName(domains[i+1]),
				},
			)
		}
	}

	domain := domains[len(domains)-1]

	switch addr.BitLen() {
	case 32:
		b.AResource(
			dnsmessage.ResourceHeader{
				Name:  dnsmessage.MustNewName(domain),
				Type:  dnsmessage.TypeA,
				Class: dnsmessage.ClassINET,
				TTL:   0,
			},
			dnsmessage.AResource{
				A: addr.As4(),
			},
		)
	case 128:
		b.AAAAResource(
			dnsmessage.ResourceHeader{
				Name:  dnsmessage.MustNewName(domain),
				Type:  dnsmessage.TypeAAAA,
				Class: dnsmessage.ClassINET,
				TTL:   0,
			},
			dnsmessage.AAAAResource{
				AAAA: addr.As16(),
			},
		)
	default:
		panic("invalid address length")
	}
	return must.Get(b.Finish())
}

func prefixEqual(a, b netip.Prefix) bool {
	return a == b
}

func prefixCompare(a, b netip.Prefix) int {
	if a.Addr().Compare(b.Addr()) == 0 {
		return a.Bits() - b.Bits()
	}
	return a.Addr().Compare(b.Addr())
}
